from pathlib import Path

import pytest

from bacpage.src import common_funcs


def test_recognize_folder_of_fastas():
    search_directory = "test/test_tree_fasta_directory"
    found = common_funcs.load_input( directory=search_directory, minimum_completeness=0 )
    search_directory = Path( search_directory ).absolute()
    expected = [search_directory / f"t{i}.fasta" for i in range( 1, 21 )]
    assert sorted( found.values() ) == sorted( expected )


def test_recognize_project_directory():
    search_directory = "test/test_tree_project_directory"
    found = common_funcs.load_input( directory=search_directory, minimum_completeness=0 )
    search_directory = Path( search_directory ).absolute() / "results/consensus"
    expected = [search_directory / f"t{i}.fasta" for i in range( 1, 21 )]
    assert sorted( found.values() ) == sorted( expected )


def test_remove_low_coverage_sequences():
    search_directory = "test/test_tree_fasta_directory"
    found = common_funcs.load_input( directory=search_directory, minimum_completeness=0.9 )
    search_directory = Path( search_directory ).absolute()
    expected = [search_directory / f"t{i}.fasta" for i in range( 1, 15 )]
    assert sorted( found.values() ) == sorted( expected )


def test_raise_error_with_duplicate_samples_in_directory():
    search_directory = "test/test_tree_duplicate_samples"
    with pytest.raises( SystemExit ) as excinfo:
        found = common_funcs.load_input( directory=search_directory, minimum_completeness=0.9 )
    assert excinfo.value.code == -5


def test_config_finds_local_parameters():
    config = common_funcs.load_configfile(
        "test/configs/local_parameters.yaml", Path( "test/test_pipeline" ).absolute(), schema="phylogeny"
    )
    reference = Path( config["reference"] )
    assert reference.is_absolute() and reference.exists(), f"Local reference in resources directory was not correctly found ({reference})."
    recombinant_mask = Path( config["recombinant_mask"] )
    assert recombinant_mask.is_absolute() and recombinant_mask.exists(), f"Local reference gene directory was not correctly found ({recombinant_mask})."



from pathlib import Path

import pytest

from bacpage.src import profiling


def test_find_folder_of_fastas():
    search_directory = Path( "test/test_tree_fasta_directory" ).absolute()
    found = profiling.find_sequences( directory=search_directory )
    search_directory = Path( search_directory ).absolute()
    expected = [search_directory / f"t{i}.fasta" for i in range( 1, 21 )]
    assert sorted( found.values() ) == sorted( expected )


def test_find_consensus_sequences():
    search_directory = Path( "test/test_tree_project_directory" ).absolute()
    found = profiling.find_sequences( directory=search_directory )
    search_directory = Path( search_directory ).absolute() / "results/consensus"
    expected = [search_directory / f"t{i}.fasta" for i in range( 1, 21 )]
    assert sorted( found.values() ) == sorted( expected )


def test_find_assemblies():
    search_directory = Path( "test/test_find_assemblies" ).absolute()
    found = profiling.find_sequences( directory=search_directory )
    search_directory = Path( search_directory ).absolute() / "results/assembly"
    expected = [search_directory / f"t{i}.annotated.gff" for i in range( 1, 10 )]
    assert sorted( found.values() ) == sorted( expected )


def test_raise_error_with_duplicate_samples_in_directory():
    search_directory = Path( "test/test_tree_duplicate_samples" ).absolute()
    with pytest.raises( SystemExit ) as excinfo:
        found = profiling.find_sequences( directory=search_directory )
    assert excinfo.value.code == -5


def test_profiling_snakemake_works():
    project_directory = Path( "test/test_profiling/" )
    expected_output = ["results/reports/antibiotic_resistance_detailed.tsv",
                       "results/reports/antibiotic_resistance.tsv"]
    expected_output = [project_directory / file for file in expected_output]
    profiling.profile_sequences(
        project_directory,
        ".",
        "card", -1,
        False
    )

    results = dict()
    for file in expected_output:
        results[file] = file.exists()
    assert all( results.values() ), f"Not all expected files generated. Expected {results}"

    expected_output[0].unlink( missing_ok=True )


def test_profile_postamble():
    project_directory = Path( "test/test_profiling/" ).absolute()
    profiling.postamble( project_directory )



import shutil
from pathlib import Path

import pytest
from Bio import AlignIO

from bacpage.src.utils import extract_region


@pytest.fixture
def extraction_run():
    regions = "test/test_extract_script/regions.bed"
    directory = "test/test_extract_script/"
    directory_path = Path( directory )

    extract_region.extract_regions( regions=regions, directory=directory )
    yield directory_path / extract_region.OUTPUT_PATH

    if (directory_path / extract_region.OUTPUT_PATH).exists():
        shutil.rmtree( directory_path / extract_region.OUTPUT_PATH )


def tests_output_directory_created( extraction_run: Path ):
    assert extraction_run.exists(), f"Output directory {extraction_run} was not created."


def test_correct_number_of_regions_extracted( extraction_run: Path ):
    files = [i for i in extraction_run.iterdir() if i.suffix == ".fasta"]
    got = len( files )
    want = 2
    assert got == want, f"Incorrect number of regions extracted. Want {want}, got {got}."


def test_results_are_valid_fastas( extraction_run: Path ):
    for file in extraction_run.iterdir():
        if file.suffix == ".fasta":
            records = AlignIO.read( file, "fasta" )


def test_correct_number_of_sequences_extracted( extraction_run: Path ):
    number_of_sequences = []
    expected_number_of_sequencs = 4
    for file in extraction_run.iterdir():
        if file.suffix == ".fasta":
            records = AlignIO.read( file, "fasta" )
            number_of_sequences.append( (file, len( records )) )
    assert all( map( lambda x: x[1] == expected_number_of_sequencs,
                     number_of_sequences ) ), f"Incorrect number of sequences in a result: {number_of_sequences}"


def test_correct_extraction_length_ctxA( extraction_run: Path ):
    ctxA = AlignIO.read( extraction_run / "ctxA.extracts.fasta", "fasta" )
    got = ctxA.get_alignment_length()
    want = 776
    assert got == want


def test_correct_extraction_length_ctxB( extraction_run: Path ):
    ctxb = AlignIO.read( extraction_run / "ctxB.extracts.fasta", "fasta" )
    got = ctxb.get_alignment_length()
    want = 374
    assert got == want



import argparse
import sys
from pathlib import Path

import snakemake

from bacpage.src import common_funcs


def add_command_arguments( parser: argparse.ArgumentParser ):
    parser.description = "Reconstructs maximum likelihood phylogeny from consensus sequences."

    parser.add_argument(
        "directory", type=str, nargs="?", default=".", help="Path to valid project directory [current directory]."
    )
    parser.add_argument(
        "--configfile", type=str, default=".", help="Path to assembly configuration file ['config.yaml']."
    )
    parser.add_argument(
        "--database", type=str, default=None, help="Database to use for antimicrobial resistance profiling."
    )
    parser.add_argument( "--threads", type=int, default=-1, help="Number of threads available for command [all]." )
    parser.add_argument( "--verbose", action="store_true", help="Print lots of stuff to screen." )

    parser.set_defaults( command=profiling_entrypoint )


def find_sequences( directory: Path ) -> dict[str, Path]:
    assert directory.exists(), f"{directory} does not exist."

    # search for assemblies
    assembly_loc = directory / "results/assembly"
    consensus_loc = directory / "results/consensus"
    extensions = [".fa", ".fasta"]
    for search_location, exts in [(assembly_loc, [".gff"]), (consensus_loc, extensions), (directory, extensions)]:
        if search_location.exists():
            files = common_funcs.find_files( search_location, exts )
            if len( files ) > 0:
                return files
    sys.stderr.write(
        f"Unable to find assemblies, consensus sequences, or loose fastas in {directory}. Please indicate a valid project directory.\n"
    )
    sys.exit( -8 )


def postamble( directory: Path ):
    print()
    print( "Successfully profiled provided samples." )
    print(
        f"A report detailing the antimicrobial resistance genes detected in each sample can be found at {directory / 'results/reports/antibiotic_resistance.tsv'}. " )


def profile_sequences( project_directory, configfile, database, threads, verbose ):
    project_path = Path( project_directory ).absolute()
    assert project_path.exists() and project_path.is_dir(), f"Specified project directory {project_path} does not exist. Please specify a valid directory."

    print( "Searching for sequences...", end="" )
    try:
        input_sequences = find_sequences( project_path )
    except:
        print( "Error" )
        raise
    print( f"Done. {len( input_sequences )} found." )

    # Check config file
    print( "Loading and validating configuration file...", end="" )
    try:
        config = common_funcs.load_configfile( configfile, project_path )
    except Exception:
        print( "Error" )
        raise
    print( "Done" )

    config["SAMPLES"] = input_sequences

    if database:
        config["antibiotic_resistance"]["database"] = database

    # Calculate number of threads
    useable_threads = common_funcs.calculate_threads( threads )

    snakefile = common_funcs.PACKAGE_DIR / "rules/profiling.smk"
    assert snakefile.exists(), f"Snakefile {snakefile} does not exist."
    status = snakemake.snakemake(
        snakefile, printshellcmds=True, forceall=True, force_incomplete=True, workdir=project_path,
        restart_times=common_funcs.RESTART_TIMES, config=config, cores=useable_threads, lock=False, quiet=not verbose
    )
    if not status:
        sys.stderr.write( "Snakemake pipeline did not complete successfully. Check for error messages and rerun.\n" )
        sys.exit( -2 )

    postamble( project_path )


def profiling_entrypoint( args: argparse.Namespace ):
    profile_sequences(
        project_directory=args.directory,
        configfile=args.configfile,
        database=args.database,
        threads=args.threads,
        verbose=args.verbose
    )

# Identify reference genes
# print( "Identifying gene sequences for typing...", end="" )
# try:
#    GENES = get_genes( config["reference_genes"] )
# except Exception:
#    print( "Error" )
#    raise
# print( "Done" )
# print( f"The following genes will be used: [{', '.join( GENES.keys() )}]\n" )



import argparse
import pandas as pd
from subprocess import run
from io import StringIO

IDX_COLUMNS = ["gene", "length", "reads_mapped", "reads_unmapped"]
COV_COLUMNS = ["gene", "position", "depth"]


def collect_reads( alignment: str ) -> pd.DataFrame:
    command = run( f"samtools idxstats {alignment}", shell=True, capture_output=True, text=True )
    reads = pd.read_csv( StringIO( command.stdout ), sep="\t", names=IDX_COLUMNS )
    reads["frac_mapped"] = reads["reads_mapped"] / (reads["reads_mapped"] + reads["reads_unmapped"]).sum()
    reads = reads.loc[reads["gene"] != "*"]
    return reads


def calculate_coverage( alignment: str, minimum_coverage: int ) -> pd.DataFrame:
    command = run( f"samtools depth -a {alignment}", shell=True, capture_output=True, text=True )
    cov = pd.read_csv( StringIO( command.stdout ), sep="\t", names=COV_COLUMNS )
    cov = cov.groupby( "gene" )["depth"].agg( ["count", "mean", lambda x: sum( x > minimum_coverage )] )
    cov.columns = ["pos_total", "mean_depth", "pos_covered"]
    cov["frac_covered"] = cov["pos_covered"] / cov["pos_total"]
    return cov


def calculate_stats( alignment: str, name: str, minimum_coverage: int, output: str ) -> None:
    reads_df = collect_reads( alignment )
    coverage = calculate_coverage( alignment, minimum_coverage=minimum_coverage )
    output_df = coverage.merge( reads_df, on="gene" )
    output_df = output_df.drop( columns=["reads_unmapped", "length"] )
    output_df["sample"] = name
    output_df = output_df[
        ["sample", "gene", "mean_depth", "pos_total", "pos_covered", "frac_covered", "reads_mapped", "frac_mapped"]]
    output_df.to_csv( output, index=False )


if __name__ == "__main__":
    parser = argparse.ArgumentParser( description="Calculates alignment stats for a given BAM file." )

    # Initialize optional arguments
    parser.add_argument( "-a", "--alignment", required=True, type=str, help="input alignment" )
    parser.add_argument( "-m", "--min-coverage", required=True, type=int,
                         help="minimum coverage required to call a base." )
    parser.add_argument( "-s", "--sample-name", required=False, default=None, help="" )
    parser.add_argument( "-o", "--output", required=True, type=str, help="location to save output" )

    args = parser.parse_args()
    sample_name = args.sample_name if args.sample_name else args.alignment
    calculate_stats( alignment=args.alignment, name=sample_name, minimum_coverage=args.min_coverage,
                     output=args.output )



import argparse
import sys
from pathlib import Path

from jinja2 import Environment, FileSystemLoader

from bacpage.src import common_funcs


def add_command_arguments( parser: argparse.ArgumentParser ):
    parser.description = "Set up project directory for analysis."
    parser.add_argument( "directory", help="Location to create project directory" )
    parser.add_argument( "--quiet", action="store_true", help="Do not display helpful messages during creation." )
    parser.add_argument( "--force", action="store_true", help="Generate directory wihtout checking if it is empty." )
    parser.set_defaults( command=example_entrypoint )


def example_entrypoint( args: argparse.Namespace ):
    create_project_directory( args.directory, quiet=args.quiet, force=args.force )


def create_project_directory( directory: str, quiet: bool = False, force: bool = False ):
    project_directory = Path( directory ).absolute()

    if project_directory.exists():
        if any( project_directory.iterdir() ) and not force:
            sys.stderr.write( f"Could not create project directory. {project_directory} must be empty.\n" )
            sys.exit( -8 )
    else:
        # create project directory
        project_directory.mkdir()

    # create input directory
    # TODO: Cleanup created files if error occured. Wrap below in a function and use try-except-finally
    (project_directory / "input").mkdir()

    # create samples file
    sample_data = project_directory / "sample_data.csv"
    # with sample_data.open( "w" ) as samples_file:
    #    samples_file.write( "sample,read1,read2\n" )
    #    samples_file.write( "a,path-to-a-read1,path-to-a-read2\n" )
    #    samples_file.write( "b,path-to-b-read1,path-to-b-read2\n" )
    #    samples_file.write( "c,path-to-c-read1,path-to-c-read2\n" )

    # create config file
    environment = Environment( loader=FileSystemLoader( common_funcs.PACKAGE_DIR / "schemas/" ) )
    template = environment.get_template( "illumina_config.template.yaml" )

    project_config = dict()
    project_config["sample_data"] = sample_data
    project_config['project_path'] = project_directory
    content = template.render( project_config )
    with (project_directory / "config.yaml").open( "w" ) as config_file:
        config_file.write( content )

    if not quiet:
        print( f"Creating project directory at {directory}." )
        print( f"The absolute path of the project directory is {project_directory}" )
        print()
        print( "The structure of the project directory is as follows:" )
        print( project_directory.name + "/" )
        print( "├── input/" )
        print( "└── config.yaml" )
        print( "" )
        print( "To assemble genomes, place raw sequencing FASTQs in the input directory." )
        print( "Then, use `bacpage assemble` to generate consensus sequences or de novo assembles." )



import argparse
from io import StringIO
from subprocess import run

from Bio import SeqIO


def convert_vcf( vcf: str, reference: str, output: str ):
    # load reference and get id
    reference_seq = SeqIO.read( reference, "fasta" )
    reference_id = reference_seq.id

    # Load samples in vcf and remove reference if present.
    query = run( f"bcftools query -l {vcf}", shell=True, capture_output=True, text=True )
    samples = query.stdout.strip().split( "\n" )
    samples = [sample for sample in samples if sample != reference_id]

    records = list()
    with open( output, "w" ) as output_file:
        for sample in samples:
            consensus = run(
                f"bcftools consensus --mark-del N -f {reference} -s '{sample}' {vcf}", shell=True, capture_output=True,
                text=True
            )
            record = SeqIO.read( StringIO( consensus.stdout ), "fasta" )
            record.id = sample
            record.name = ""
            record.description = ""
            record.seq = record.seq.replace( "*", "N" ).upper()
            SeqIO.write( record, output_file, "fasta" )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Converts a multi-sample VCF file to a multisequence alignment in FASTA format."
    )

    # Initialize optional arguments
    parser.add_argument( "--vcf", type=str, help="VCF file to convert" )
    parser.add_argument( "--reference", type=str, help="Reference sequence used when generating the VCF." )
    parser.add_argument( "--output", type=str, help="Location to save fasta file." )
    args = parser.parse_args()

    convert_vcf( args.vcf, args.reference, args.output )



import argparse
import sys
from pathlib import Path


def add_command_arguments( parser: argparse.ArgumentParser ):
    parser.description = "Generate a valid sample_data.csv from a directory of FASTQs."
    parser.add_argument(
        "directory", default=".", help="location of FASTQ files [current directory]"
    )
    parser.add_argument(
        "--delim", default="_", type=str, help="deliminator to extract sample name from file name [_]"
    )
    parser.add_argument(
        "--index", default=0, type=int, help="index of sample name after splitting file name by delim [0]"
    )
    parser.add_argument(
        "--output", type=str, default="sample_data.csv", help="location to save sample data ['sample_data.csv']"
    )

    parser.set_defaults( command=identify_entrypoint )


def write_samples_to_file( sample_data: dict[str, list[str]], output: Path ):
    with open( output, "w" ) as output_file:
        output_file.write( "sample,read1,read2\n" )
        for sample in sample_data:
            files = sorted( sample_data[sample] )
            try:
                output_file.write( f"{sample},{files[0]},{files[1]}\n" )
            except IndexError:
                sys.stderr.write( f"Unable to find paired sequencing reads for {sample}. Only {files[0]} is present." )
                sys.exit( -1 )


def identify_entrypoint( args: argparse.Namespace ):
    sample_data = generate_sample_data( directory=args.directory, delim=args.delim, index=args.index )
    write_samples_to_file( sample_data=sample_data, output=args.output )
    postamble( files=sample_data, directory=Path( args.directory ), output=Path( args.output ) )


def generate_sample_data( directory: str, delim: str = "_", index: int = 0 ) -> dict[str, list[str]]:
    directory = Path( directory )
    samples = dict()
    for file in directory.iterdir():
        if file.name.endswith( ("fastq.gz", "fq.gz") ):
            sample_name = file.name.split( delim )[index]
            if sample_name in samples:
                samples[sample_name].append( file.absolute() )
            else:
                samples[sample_name] = [file.absolute()]
    for sample in samples:
        samples[sample] = sorted( samples[sample] )

    return samples


def postamble( files: dict[str, list[str]], directory: Path, output: Path ):
    print()
    print( f"Identified {len( files )} samples in {directory}." )
    print( f"Saving the location of these files to {output}." )
    print( "This project directory is now set up for assembly using `bacpage assemble`." )



import argparse
import sys
from pathlib import Path

from Bio import SeqIO

from bacpage.src import common_funcs

OUTPUT_PATH = "results/extractions/"


def add_command_arguments( parser: argparse.ArgumentParser ):
    parser.description = "Extract regions specified in a bed file from consensus sequences in a project directory."
    parser.add_argument(
        "directory", default=".", help="location of FASTQ files [current directory]"
    )
    parser.add_argument( "--region", help="BED file containing region(s) to extract." )
    parser.add_argument( "--output", help="Directory to save extraction alignments" )

    parser.set_defaults( command=extract_entrypoint )


def parse_bed_file( region ):
    region_path = Path( region ).absolute()

    if not region_path.exists():
        sys.stderr.write( f"{region_path} does not exist. Please specify a valid BED file with the '--region' option." )
        sys.exit( -1 )

    parsed_regions = dict()
    with open( region_path, "r" ) as region_file:
        for line in region_file:
            fields = line.strip().split( "\t" )
            try:
                name = fields[0]
                start = int( fields[1] )
                end = int( fields[2] )
            except IndexError:
                sys.stderr.write(
                    f"Unable to parse line: '{line}' in bed file due to not enough fields. Make sure {region_file} contains at least three fields: name, start, and end"
                )
                sys.exit( -2 )

            parsed_regions[name] = [start, end]
    return parsed_regions


def extract_regions( regions: str, directory: str ):
    # Load bed file
    bed = parse_bed_file( regions )
    region_names = list( bed.keys() )

    # Load fasta files
    input_sequences = common_funcs.load_input( directory=directory, minimum_completeness=0 )

    # Iterate through regions
    extractions = {name: [] for name in region_names}
    for sample, sequence_file in input_sequences.items():
        sequence = SeqIO.read( sequence_file, "fasta" )
        for name, region in bed.items():
            extraction = sequence[region[0]:region[1]]
            extractions[name].append( extraction )

    for name, extracts in extractions.items():
        outdir_path = Path( directory ).absolute() / OUTPUT_PATH
        outdir_path.mkdir( exist_ok=True )
        SeqIO.write( extracts, outdir_path / f"{name}.extracts.fasta", "fasta" )

    postamble( regions=region_names, directory=Path( directory ) )


def extract_entrypoint( args: argparse.Namespace ):
    extract_regions(
        regions=args.region,
        directory=args.directory
    )


def postamble( regions: list[str], directory: Path ):
    print()
    print( f"Extracted regions {', '.join( regions )} from files in {directory}." )
    print( f"Multi-sequence alignments for each regions have been saved to {directory / OUTPUT_PATH}" )



import argparse

from bacpage import __version__


def print_version( args: argparse.Namespace ):
    print( __version__ )


def add_command_arguments( parser: argparse.ArgumentParser ):
    parser.set_defaults( command=print_version )



import argparse
import sys
from pathlib import Path
from subprocess import run

import snakemake
from Bio import SeqIO

from bacpage.src import common_funcs


def add_command_arguments( parser: argparse.ArgumentParser ):
    parser.description = "Reconstructs maximum likelihood phylogeny from consensus sequences."

    parser.add_argument(
        "directory", type=str, default=".", help="Path to valid project directory [current directory]."
    )
    parser.add_argument(
        "--configfile", type=str, default=".", help="Path to assembly configuration file ['config.yaml']."
    )
    parser.add_argument(
        "--minimum-completeness", type=float, default=0.9,
        help="minimum coverage required to be included in analysis [0.9]"
    )
    parser.add_argument(
        "--terra", action="store_true",
        help="Generate input for Terra and print instructions. [False; runs pipeline locally]"
    )
    parser.add_argument(
        "--detect", type=str,
        help="Bypass computationally intensive recombination detection by supplying a GFF file previous generated by gubbins."
    )
    parser.add_argument(
        "--no-detect", action="store_true",
        help="Skip performing recombinant region detection [False; perform detection]"
    )
    parser.add_argument(
        "--mask", type=str, default=None,
        help="gff file used to mask to all sequences prior to tree building. If not specified, sequences will not be masked [False]"
    )
    parser.add_argument( "--threads", type=int, default=-1, help="Number of threads available for command [all]." )
    parser.add_argument( "--verbose", action="store_true", help="Print lots of stuff to screen." )

    parser.set_defaults( command=phylogeny_entrypoint )


def parse_names_fasta( dataset: Path ):
    return_list = []
    if dataset and (dataset != ""):
        for record in SeqIO.parse( dataset, "fasta" ):
            return_list.append( record.name )
    return return_list


def parse_names_vcf( dataset: Path ):
    if dataset.suffix == ".gz":
        run( f"bcftools index {dataset}", shell=True )
    response = run( f"bcftools query -l {dataset}", shell=True, capture_output=True,
                    text=True )
    names = response.stdout.strip().split( "\n" )
    return names


def validate_sequences( sequence_paths: dict[str, Path], reference: str, background_dataset: Path = None ):
    seen_names = list()

    for path in sequence_paths.values():
        for record in SeqIO.parse( path, "fasta" ):
            seen_names.append( record.name )

    if background_dataset and background_dataset != "":
        if background_dataset.suffix in [".fa", ".fasta", ".fsa"]:
            seen_names.extend( parse_names_fasta( background_dataset ) )
        elif background_dataset.name.endswith( (".vcf", ".vcf.gz") ):
            seen_names.extend( parse_names_vcf( background_dataset ) )
        else:
            sys.stderr.write(
                f"Unable to automatically detect format of {background_dataset}. Please rename with common FASTA or VCF extension."
            )
            sys.exit( -1 )

    if len( seen_names ) != len( set( seen_names ) ):
        sys.stderr.write( "Sequence names are not unique. Remove duplicate sequences.\n" )
        sys.exit( -6 )

    reference_name = next( SeqIO.parse( reference, "fasta" ) ).id
    if reference_name in seen_names:
        sys.stderr.write(
            "Reference cannot be present in samples or background dataset. Please rename reference or remove it from the background."
        )
        sys.exit( -1 )


def postamble( directory: Path ):
    print()
    print( f"Reconstructed phylogeny is available at {directory / 'results/phylogeny/phylogeny.tree'}" )
    print(
        "Open the file in any tree viewer to visualize. We recommend figtree (http://tree.bio.ed.ac.uk/software/figtree/)."
    )


def reconstruct_phylogeny( project_directory: str, configfile: str, minimum_completeness: float = 0,
                           terra: bool = False, mask: str = None, detect_file: str = None, skip_detect: bool = False,
                           threads: int = 1, verbose: bool = True, dryrun: bool = False ):
    project_path = Path( project_directory ).absolute()

    if not project_path.exists() or not project_path.is_dir():
        sys.stderr.write(
            f"Specified project directory {project_path} does not exist. Please specify a valid directory.\n"
        )
        sys.exit( -1 )

    # Check config file
    print( "Loading and validating configuration file...", end="" )
    try:
        config = common_funcs.load_configfile( configfile, project_path, schema="phylogeny" )
    except Exception:
        print( "Error" )
        raise
    print( "Done" )

    if config["background_dataset"] not in ["", "<background-dataset-path>"]:
        background_dataset = Path( config["background_dataset"] ).absolute()
        config["BACKGROUND"] = background_dataset
    else:
        config["BACKGROUND"] = ""

    print( "Loading and validating input sequences...", end="" )
    input_sequences = common_funcs.load_input( project_directory, minimum_completeness=minimum_completeness )
    validate_sequences( input_sequences, config["reference"], config["BACKGROUND"] )
    print( f"Done. Found {len( input_sequences )} sequences in directory." )

    config["SAMPLES"] = input_sequences

    # Add command line arguments
    config["TERRA"] = terra
    config["DETECT"] = not skip_detect
    if not skip_detect:
        if detect_file:
            config["DETECT"] = detect_file
    config["MASK"] = False
    config["MASK_LOCATION"] = ""
    if mask:
        mask_loc = Path( mask ).absolute()
        if not mask_loc.exists():
            sys.stderr.write(
                f"{mask_loc}, specified on command line, does not exist. Please specify a valid GFF file."
            )
            sys.exit( -1 )
        config["MASK"] = True
        config["MASK_LOCATION"] = str( mask_loc )
    # elif config["recombinant_mask"] != "":
    #    mask_loc = Path( config["recombinant_mask"] ).absolute()
    #    if not mask_loc.exists():
    #        sys.stderr.write(
    #            f"{mask_loc}, specified in config file, does not exist. Please specify a valid GFF file or leave blank." )
    #        sys.exit( -1 )
    #    config["MASK"] = True
    #    config["MASK_LOCATION"] = str( mask_loc )

    # Calculate number of threads
    useable_threads = common_funcs.calculate_threads( threads )

    # load snakefile
    snakefile = common_funcs.PACKAGE_DIR / "rules/phylogeny.smk"
    assert snakefile.exists(), f"Snakefile does not exist. Searching for {snakefile}"

    if dryrun:
        return config, snakefile

    status = snakemake.snakemake(
        snakefile, printshellcmds=True, forceall=True, force_incomplete=True, workdir=project_directory,
        restart_times=common_funcs.RESTART_TIMES, config=config, cores=useable_threads, lock=False, quiet=not verbose
    )
    if not status:
        sys.stderr.write( "Snakemake pipeline did not complete successfully. Check for error messages and rerun.\n" )
        sys.exit( -2 )

    postamble( project_path )


def phylogeny_entrypoint( args: argparse.Namespace ):
    reconstruct_phylogeny(
        project_directory=args.directory,
        configfile=args.configfile,
        minimum_completeness=args.minimum_completeness,
        terra=args.terra,
        mask=args.mask,
        detect_file=args.detect,
        skip_detect=args.no_detect,
        threads=args.threads,
        verbose=args.verbose,
    )



import sys
from importlib.resources import files
from pathlib import Path

import yaml
from Bio import SeqIO
from snakemake import WorkflowError
from snakemake.utils import validate

DEFAULT_CONFIG = "config.yaml"
DEFAULT_SAMPLEDATA = "sample_data.csv"
PACKAGE_DIR = files( "bacpage" )
CONFIG_PATHS = {"Illumina": ["reference"], "phylogeny": ["reference", "recombinant_mask"], "assemble": ["reference"]}
RESTART_TIMES = 0
OTHER_IUPAC = {'r', 'y', 's', 'w', 'k', 'm', 'd', 'h', 'b', 'v'}
VALID_CHARACTERS = [{'a'}, {'c'}, {'g'}, {'t'}, {'n'}, OTHER_IUPAC, {'-'}, {'?'}]


def find_files( directory: Path, extensions: list[str] ) -> dict[str, Path]:
    search_path = directory.absolute()
    print( search_path )
    found = dict()
    for file in search_path.iterdir():
        if file.suffix in extensions:
            name = file.stem
            if name in found:
                sys.stderr.write( f"{name} was found in two or more files. Duplicate sequences cannot be processed.\n" )
                sys.exit( -5 )
            path = search_path / file
            found[name] = path
    return found


def is_project_directory( directory ):
    return (directory / "sample_data.csv").exists() and (directory / "config.yaml").exists()


def normalize_path( value: str, working_directory: Path ) -> Path:
    path = Path( value )
    if not path.is_absolute():
        return working_directory / path
    else:
        return path


def load_configfile( specified_loc: str, project_directory: Path, schema: str = "Illumina" ) -> dict:
    """ Attempts for find config file using user supplied information. If config file is directly specified, use it, else
    search for the config file in the project directory.

    Parameters
    ----------
    specified_loc: str
        Path to config file. Pass "." to automatically search for file in project directory.
    project_directory: pathlib.Path
        Path to project directory. Used if config file path is not specified and to normalize relative paths in the config file.
    schema: str
        Specify a different schema to use to validate configuration file.

    Returns
    -------
    dict
        Config file loaded as a python object.
    """
    configfile_loc = Path( specified_loc ).absolute()
    if specified_loc == ".":
        configfile_loc = project_directory / DEFAULT_CONFIG
        assert configfile_loc.exists(), "Unable to automatically find config in project directory (searching for 'config.yaml'). Please specify a valid configuration file."
    assert configfile_loc.exists(), f"{configfile_loc} does not exist. Please specify a valid file."

    with open( configfile_loc, "r" ) as cf:
        configfile = yaml.safe_load( cf )

    schema_location = PACKAGE_DIR / f"schemas/{schema}_config.schema.yaml"

    try:
        validate( configfile, schema_location )
    except WorkflowError as err:
        # message = err.args[0].split( "\n" )
        print( "Error" )
        # sys.stderr.write( f"{message[0]} {message[1].split( ': ' )[1]}.\n" )
        sys.stderr.write( err.args[0] )
        sys.exit( -9 )

    config_paths = CONFIG_PATHS.get( schema, [] )
    for key in config_paths:
        configfile[key] = str( normalize_path( configfile[key], PACKAGE_DIR / "resources" ) )

    return configfile


def calculate_threads( threads ):
    if threads == 0:
        sys.stderr.write(
            "Pipeline cannot function without threads. Please specify a non-zero number of threads with the '--threads' "
            "option or run comman without '--threads' option to automatically detected the number of available threads.\n"
        )
        sys.exit( -4 )
    elif threads < 0:
        import multiprocessing
        threads = multiprocessing.cpu_count()
    print( f"Using {threads} threads." )

    return threads


def calculate_completeness( sequence_loc: Path ) -> float:
    record = SeqIO.read( sequence_loc, "fasta" )
    seq = record.seq.lower()
    l = len( seq )
    counts = []

    for v in VALID_CHARACTERS:
        counts.append( sum( map( lambda x: seq.count( x ), v ) ) )
    invalid_nucleotides = l - sum( counts )

    if invalid_nucleotides > 0:
        print( "Invalid characters in sequence. Might not be a valid nucleotide sequence." )

    return 1.0 - (counts[4] / l)


def load_input( directory: str, minimum_completeness: float ) -> dict[str, Path]:
    """ Searches the indicated directory for fasta files. Specifically, it will search in 'results/consensus' if the
    directory is identified as a project directory, otherwise the root directory. Additionally, the function can filter
    out fasta files which contain more than 'minimum_completeness' ambiguous characters.

    Parameters
    ----------
    directory: str
        Directory to search within. Can be a project directory.
    minimum_completeness: float
        Proportion of non-ambiguous characters a fasta file must have to be returned. Set to 0 to return all fasta
        files.

    Returns
    -------
    dict[str, Path]
        Dictionary containing sample names mapped to absolute path for each fasta file found in 'directory.'
    """
    search_directory = Path( directory ).absolute()
    if is_project_directory( search_directory ):
        search_directory = search_directory / "results/consensus"

    fastas = find_files( directory=search_directory, extensions=[".fa", ".fasta"] )

    if minimum_completeness > 0:
        for name in list( fastas.keys() ):
            completeness = calculate_completeness( fastas[name] )
            if completeness < minimum_completeness:
                del fastas[name]
    return fastas



import shutil
from contextlib import redirect_stdout
from io import StringIO
from pathlib import Path
from typing import Any

import pandas as pd
import pytest
import snakemake
from Bio import Phylo

from bacpage.src import phylogeny


def test_error_if_vcf_background_doesnt_match_reference():
    search_directory = "test/test_tree_vcf"
    with pytest.raises( SystemExit ) as excinfo:
        phylogeny.reconstruct_phylogeny(
            project_directory=search_directory,
            configfile=".",
            minimum_completeness=0,
            threads=1,
            verbose=True
        )
    assert excinfo.value.code < 0


def test_raise_error_if_reference_present_already():
    search_directory = "test/test_tree_project_directory"
    with pytest.raises( SystemExit ) as excinfo:
        phylogeny.reconstruct_phylogeny(
            project_directory=search_directory,
            configfile="test/test_tree_project_directory/duplicate_reference.yaml",
            minimum_completeness=0,
            threads=1,
            verbose=True,
            dryrun=True
        )
    assert excinfo.value.code < 0


def test_raise_error_with_duplicate_sample_names():
    search_directory = "test/test_tree_duplicate_sequences"
    with pytest.raises( SystemExit ) as excinfo:
        phylogeny.reconstruct_phylogeny(
            project_directory=search_directory,
            configfile="test/test_tree_fasta_directory/config.yaml",
            minimum_completeness=0,
            threads=1,
            verbose=True
        )
    assert excinfo.value.code == -6


def test_parse_vcf_names():
    vcf = Path( "test/test_vcf.vcf" )
    found = phylogeny.parse_names_vcf( vcf )
    expected = [
        "Africa|TZA|SAMN19110433|T10|2017",
        "Africa|TZA|SAMN19110437|T13|2016",
        "Africa|TZA|SAMN19110438|T13|2016",
        "Africa|TZA|SAMN19110441|T13|2017",
        "Africa|TZA|SAMN19110439|T13|2016",
        "Africa|TZA|SAMN19110428|T13|2017",
        "Africa|TZA|SAMN19110430|T13|2017",
        "Africa|TZA|SAMN19110431|T13|2017",
        "Africa|TZA|SAMN19110436|T13|2017",
        "Africa|TZA|SAMN19110443|T13|2017"
    ]
    assert sorted( found ) == sorted( expected )


def test_vcf_to_fasta():
    from bacpage.scripts.vcf_to_fasta import convert_vcf
    import filecmp
    vcf_file = "test/test_vcf_to_fasta/input.vcf.gz"
    reference = "test/test_vcf_to_fasta/reference.fasta"
    output = "test/test_vcf_to_fasta/output.fasta"

    convert_vcf( vcf=vcf_file, reference=reference, output=output )
    assert Path( output ).exists() and Path( output ).is_file(), f"{output} does not exist or is not a file."
    assert filecmp.cmp( "test/test_vcf_to_fasta/input.fasta", output, shallow=True )


def get_rules_dryrun( snakefile: Path, config: dict[str, Any], workdir: str ):
    with redirect_stdout( StringIO() ) as f:
        value = snakemake.snakemake( snakefile, forceall=True, workdir=workdir, config=config, summary=True )
    assert value, "Snakemake file is not valid"
    df = pd.read_csv( StringIO( f.getvalue() ), sep="\t" )
    return df["rule"].unique()


def test_correct_rules_run_if_fasta_background():
    search_directory = "test/test_tree_project_directory"
    config, snakefile = phylogeny.reconstruct_phylogeny(
        project_directory=search_directory,
        configfile="test/test_tree_project_directory/background_fasta.yaml",
        minimum_completeness=0,
        threads=1,
        verbose=True,
        dryrun=True,
    )
    estimated_rules = get_rules_dryrun( snakefile, config, search_directory )
    expected_rules = ["concatenate_sequences", "concatenate_reference", "convert_to_vcf", "generate_alignment_from_vcf",
                      "sparsify_alignment", "run_gubbins", "generate_tree", "move_tree_and_rename",
                      "move_recombinant_mask"]

    assert sorted( estimated_rules ) == sorted( expected_rules )


def test_correct_rules_run_if_vcf_background():
    search_directory = "test/test_tree_project_directory"
    config, snakefile = phylogeny.reconstruct_phylogeny(
        project_directory=search_directory,
        configfile="test/test_tree_project_directory/vcf_background.yaml",
        minimum_completeness=0,
        threads=1,
        verbose=True,
        dryrun=True,
    )
    estimated_rules = get_rules_dryrun( snakefile, config, search_directory )
    expected_rules = ["concatenate_sequences", "concatenate_reference", "convert_to_vcf", "index_background_vcf",
                      "combine_sequences_and_background_vcf", "generate_alignment_from_vcf", "run_gubbins",
                      "sparsify_alignment", "generate_tree", "move_tree_and_rename", "move_recombinant_mask"]

    assert sorted( estimated_rules ) == sorted( expected_rules )


def test_correct_rules_run_if_masking_specified():
    search_directory = "test/test_tree_project_directory"
    config, snakefile = phylogeny.reconstruct_phylogeny(
        project_directory=search_directory,
        configfile=".",
        minimum_completeness=0,
        threads=1,
        mask="bacpage/resources/cholera_mask.gff",
        verbose=True,
        dryrun=True,
    )
    estimated_rules = get_rules_dryrun( snakefile, config, search_directory )
    expected_rules = ["concatenate_sequences", "concatenate_reference", "convert_to_vcf", "convert_gff_to_bed",
                      "mask_vcf",
                      "generate_alignment_from_vcf", "run_gubbins", "sparsify_alignment", "generate_tree",
                      "move_recombinant_mask", "move_tree_and_rename"]

    assert sorted( estimated_rules ) == sorted( expected_rules )


def test_correct_rules_run_if_recombinant_masking_skipped():
    search_directory = "test/test_tree_project_directory"
    config, snakefile = phylogeny.reconstruct_phylogeny(
        project_directory=search_directory,
        configfile=".",
        minimum_completeness=0,
        threads=1,
        skip_detect=True,
        verbose=True,
        dryrun=True,
    )
    estimated_rules = get_rules_dryrun( snakefile, config, search_directory )
    expected_rules = ["concatenate_sequences", "concatenate_reference", "convert_to_vcf",
                      "generate_alignment_from_vcf", "sparsify_alignment", "generate_tree",
                      "move_tree_and_rename"]

    assert sorted( estimated_rules ) == sorted( expected_rules )


def test_correct_rules_run_if_terra_specified():
    search_directory = "test/test_tree_project_directory"
    config, snakefile = phylogeny.reconstruct_phylogeny(
        project_directory=search_directory,
        configfile=".",
        minimum_completeness=0,
        threads=1,
        terra=True,
        verbose=True,
        dryrun=True,
    )
    estimated_rules = get_rules_dryrun( snakefile, config, search_directory )
    expected_rules = ["concatenate_sequences", "concatenate_reference", "convert_to_vcf"]

    assert sorted( estimated_rules ) == sorted( expected_rules )


def test_correct_rules_run_if_gubbins_mask_supplied():
    search_directory = "test/test_tree_project_directory"
    mask = Path( "bacpage/resources/cholera_gubbins_mask.gff" ).absolute()
    config, snakefile = phylogeny.reconstruct_phylogeny(
        project_directory=search_directory,
        configfile=".",
        minimum_completeness=0,
        threads=1,
        detect_file=str( mask ),
        verbose=True,
        dryrun=True,
    )
    estimated_rules = get_rules_dryrun( snakefile, config, search_directory )
    expected_rules = ["concatenate_sequences", "concatenate_reference", "convert_to_vcf",
                      "bypass_gubbins", "generate_alignment_from_bypassed_gubbins", "sparsify_alignment",
                      "generate_tree", "move_tree_and_rename"]
    assert sorted( estimated_rules ) == sorted( expected_rules )


def test_abridged_config_accepted():
    search_directory = "test/test_tree_project_directory"
    config = "test/configs/phylogeny_only.yaml"
    config, snakefile = phylogeny.reconstruct_phylogeny(
        project_directory=search_directory,
        configfile=config,
        dryrun=True
    )
    assert config is not None


@pytest.fixture
def phylogeny_run( scope="session" ):
    project_directory = Path( "test/test_tree_fasta_directory" )
    phylogeny.reconstruct_phylogeny(
        str( project_directory ), ".", minimum_completeness=0.9, threads=-1,
        verbose=False
    )
    yield project_directory

    if (project_directory / "results").exists():
        shutil.rmtree( project_directory / "results" )
    if (project_directory / "intermediates").exists():
        shutil.rmtree( project_directory / "intermediates" )


def test_phylogeny_postamble():
    project_directory = Path( "test/test_tree_fasta_directory" ).absolute()
    phylogeny.postamble( project_directory )


@pytest.mark.slow
def test_tree_reconstruction_successfully( phylogeny_run ):
    tree = phylogeny_run / "results/phylogeny/phylogeny.tree"
    assert tree.exists() and tree.is_file(), "Phylogeny was either not created or is not a file."


@pytest.mark.slow
def test_tree_reconstruction_all_taxa_present( phylogeny_run ):
    tree = phylogeny_run / "results/phylogeny/phylogeny.tree"
    tree = Phylo.read( tree, "newick" )

    found = [clade.name for clade in tree.get_terminals()]
    expected = [f"t{i}" for i in range( 1, 15 )]

    assert sorted( found ) == sorted( expected ), "Not all taxa where found in tree."
    assert 1 < tree.total_branch_length() < 100, f"Branch length ({tree.total_branch_length()}) is not a reasonable magnitude."



import argparse
import sys

from .src import assemble, example, identify, phylogeny, profiling, version
from .src.utils import extract_region

COMMANDS = {
    "assemble"      : [assemble, "Assembles consensus sequence from raw sequencing reads."],
    "setup"         : [example, "Set up project directory for analysis."],
    "identify_files": [identify, "Generate a valid sample_data.csv from a directory of FASTQs."],
    "phylogeny"     : [phylogeny, "Align sequences and construct a maximum likelihood tree."],
    "profile"       : [profiling, "Classify consensus sequences based on the presence or absense of various genes."],
    "version"       : [version, "Prints the version of bacpage and exits."]
    # "submit" : [submit, "Prepare files for submission to online repositories."],
}

UTILITIES = {
    "extract_region": [extract_region, "Extract substring(s) from all consensus sequences."]
}


def main( sysargs=sys.argv[1:] ):
    parser = argparse.ArgumentParser(
        prog="bacpage",
        description="""██████╗  █████╗  ██████╗██████╗  █████╗  ██████╗ ███████╗
██╔══██╗██╔══██╗██╔════╝██╔══██╗██╔══██╗██╔════╝ ██╔════╝
██████╔╝███████║██║     ██████╔╝███████║██║  ███╗█████╗
██╔══██╗██╔══██║██║     ██╔═══╝ ██╔══██║██║   ██║██╔══╝
██████╔╝██║  ██║╚██████╗██║     ██║  ██║╚██████╔╝███████╗
╚═════╝ ╚═╝  ╚═╝ ╚═════╝╚═╝     ╚═╝  ╚═╝ ╚═════╝ ╚══════╝

    A bioinformatics toolkit to assemble and analyze BACterial PAthogen GEnomes""",
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    subparsers = parser.add_subparsers(
        title="Available commands",
        description="One of the following commands must be specified:",
        required=True
    )

    for command, values in COMMANDS.items():
        parser_subcommand = subparsers.add_parser( command, help=values[1] )
        values[0].add_command_arguments( parser_subcommand )

    util_command = subparsers.add_parser( "utilities", help="Miscellaneous tools." )
    util_subparser = util_command.add_subparsers(
        dest="command",
        title="Available utilities",
        description="One of the following utilities must be specified:"
    )
    for command, values in UTILITIES.items():
        util_subcommand = util_subparser.add_parser( command, help=values[1] )
        values[0].add_command_arguments( util_subcommand )

    if len( sysargs ) < 1:
        parser.print_help()
        sys.exit( -1 )
    else:
        args = parser.parse_args( sysargs )

    if not args.command:
        util_command.print_help()
        sys.exit( -1 )

    args.command( args )


if __name__ == "__main__":
    main()
